@@ -80,6 +80,8 @@ static SkkTransition okuri_nasi_transitions[] =
     { SKK_INPUT_MODE_HIRAGANA, "N A N A", "▼な*んあ【】", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "N A N a", "▼な*な【】", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "A o SPC Control_L", "▼青", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "K a k k o r y a k u SPC", "▼(略)", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "O n n o f u SPC", "▼オン/オフ", "", SKK_INPUT_MODE_HIRAGANA },
     { 0, NULL }
   };
 
@@ -121,6 +123,11 @@ static SkkTransition abort_transitions[] =
     { SKK_INPUT_MODE_HIRAGANA, "O K i C-g", "▽おき", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "O K C-g", "", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "A o i O C-g", "▽あおいお", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ m a i l SPC", "▼メイル", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ m a i l SPC C-g", "▽mail", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ m a i l SPC C-g SPC", "▼メイル", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ m a i l SPC C-g e r", "▽mailer", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ m a i l SPC C-g e r SPC", "▼メイラー", "", SKK_INPUT_MODE_HIRAGANA },
     { 0, NULL }
   };
 
@@ -166,6 +173,18 @@ static SkkTransition completion_transitions[] =
     { SKK_INPUT_MODE_HIRAGANA, "K o u k o \t \t", "▽こうこく", "", SKK_INPUT_MODE_HIRAGANA },
     /* no match for midasi word (= "あぱ") */
     { SKK_INPUT_MODE_HIRAGANA, "A p a \t", "▽あぱ", "", SKK_INPUT_MODE_HIRAGANA },
+    /* file dict has midasi word (= あい) while user dict does not */
+    { SKK_INPUT_MODE_HIRAGANA, "A i SPC C-j A \t \t", "▽あいさつ", "愛", SKK_INPUT_MODE_HIRAGANA },
+
+    /* Abbrev mode */
+    /* midasi "mail" exists */
+    { SKK_INPUT_MODE_HIRAGANA, "/ m a i \t", "▽mail", "",  SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ m a i l \t", "▽mailer", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ m a i l \t \t", "▽mailing", "", SKK_INPUT_MODE_HIRAGANA },
+    /* no more match for "mail" */
+    { SKK_INPUT_MODE_HIRAGANA, "/ m a i l \t \t \t ", "▽mailing", "", SKK_INPUT_MODE_HIRAGANA },
+    /* no match for midasi "mailingl" */
+    { SKK_INPUT_MODE_HIRAGANA, "/ m a i l i n g l \t ", "▽mailingl", "", SKK_INPUT_MODE_HIRAGANA },
     { 0, NULL }
   };
 
@@ -187,7 +206,7 @@ static SkkTransition abbrev_transitions[] =
     /* Issue#24 */
     { SKK_INPUT_MODE_HIRAGANA, "/ t e s t C-j", "", "test", SKK_INPUT_MODE_HIRAGANA },
     /* Pull request#39 */
-    { SKK_INPUT_MODE_HIRAGANA, "/ t e s t C-m", "▽test", "test", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ t e s t C-m", "", "test", SKK_INPUT_MODE_HIRAGANA },
     { 0, NULL }
   };
 
@@ -217,6 +236,13 @@ static SkkTransition dict_edit_transitions[] =
     /* Pull request#41 */
     { SKK_INPUT_MODE_HIRAGANA, "K a n j i k a t a k a n a k a n j i SPC K a n j i SPC K a t a k a n a q K a n j i SPC C-j", "▼かんじかたかなかんじ【漢字カタカナ漢字】", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "T e s u t o t e s u t o t e s u t o t e s u t o SPC t e s u t o T e s u t o q q T e s u t o q T e s u t o C-q  C-q T e s u t o q", "▼てすとてすとてすとてすと【てすとテストてすとﾃｽﾄてすと】", "", SKK_INPUT_MODE_HANKAKU_KATAKANA},
+    /* Issue#77 */
+    { SKK_INPUT_MODE_HIRAGANA, "G u g u R u G u g u q \n", "", "ググる", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "G u g u R u", "▼ググる", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "G u g u R a", "▼ググら", "", SKK_INPUT_MODE_HIRAGANA },
+    /* Issue#79 */
+    { SKK_INPUT_MODE_HIRAGANA, "H o g e f u g a SPC H o g e q F u g a SPC F u g a q RET", "▼ほげふが【ホゲフガ】", "", SKK_INPUT_MODE_HIRAGANA },
+
     { 0, NULL }
   };
 
@@ -251,6 +277,35 @@ static SkkTransition numeric_transitions[] = {
     { 0, NULL }
   };
 
+// Default insertion rule should not be activated when any modifiers are active.
+// For example, `C-T` should be unhandled (instead of inserting `T`)
+// if no command is assigned.
+static SkkTransition mods_deactivate_default_insert_transitions[] = {
+    { SKK_INPUT_MODE_HIRAGANA, "C-t", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "C-T", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "A-t", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "A-T", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "M-t", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "M-T", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "G-t", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "G-T", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "k C-t a", "", "か", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "k C-T a", "", "か", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "K C-t a", "▽か", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "K C-T a", "▽か", "", SKK_INPUT_MODE_HIRAGANA },
+    { 0, NULL }
+  };
+
+static SkkTransition commit_preedit_transitions[] = {
+    { SKK_INPUT_MODE_HIRAGANA, "A C-m", "", "あ", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "A C-j", "", "あ", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "t C-m", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "t C-j", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "n C-m", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "n C-j", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { 0, NULL }
+  };
+
 struct _SkkFixture {
   SkkContext *context;
 };
@@ -396,6 +451,58 @@ surrounding (void) {
   destroy_context (context);
 }
 
+static void
+request_selection_text_cb (SkkContext* self,
+                           gpointer   user_data)
+{
+  skk_context_set_selection_text(self, "test message");
+}
+
+static void
+selection (void) {
+  SkkContext *context = create_context (TRUE, TRUE);
+  g_signal_connect (context, "request-selection-text",
+                    G_CALLBACK (request_selection_text_cb), NULL);
+  const gchar *preedit;
+  GError *error;
+  SkkRule *rule;
+  error = NULL;
+  SkkTransition transitions[] = {
+    { SKK_INPUT_MODE_HIRAGANA, "/ C-y", "▽test message", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ C-y C-g", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ C-y C-j", "", "test message", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ t e s t t e x t SPC C-y", "▼testtext【test message】", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ t e s t t e x t SPC C-y C-g", "▽testtext", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ t e s t t e x t SPC C-y C-j", "", "test message", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "/ t e s t t e x t SPC C-y C-m", "", "test message", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "A a a a SPC C-y", "▼ああああ【test message】", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_KATAKANA, "A a a a SPC C-y", "▼アアアア【test message】", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HANKAKU_KATAKANA, "A a a a SPC C-y", "▼ｱｱｱｱ【test message】", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "A a a a SPC C-y C-m", "", "test message", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "A a a a SPC C-y RET", "", "test message", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_KATAKANA, "A a a a SPC C-y C-m", "", "test message", SKK_INPUT_MODE_KATAKANA },
+    { SKK_INPUT_MODE_KATAKANA, "A a a a SPC C-y RET", "", "test message", SKK_INPUT_MODE_KATAKANA },
+    { SKK_INPUT_MODE_HANKAKU_KATAKANA, "A a a a SPC C-y C-m", "", "test message", SKK_INPUT_MODE_HANKAKU_KATAKANA },
+    { SKK_INPUT_MODE_HANKAKU_KATAKANA, "A a a a SPC C-y RET", "", "test message", SKK_INPUT_MODE_HANKAKU_KATAKANA },
+    { SKK_INPUT_MODE_HIRAGANA, "C-y", "", "test message", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "Q C-y", "▽test message", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_KATAKANA, "C-y", "", "test message", SKK_INPUT_MODE_KATAKANA },
+    { SKK_INPUT_MODE_KATAKANA, "Q C-y", "▽test message", "", SKK_INPUT_MODE_KATAKANA },
+    { SKK_INPUT_MODE_HANKAKU_KATAKANA, "C-y", "", "test message", SKK_INPUT_MODE_HANKAKU_KATAKANA },
+    { SKK_INPUT_MODE_HANKAKU_KATAKANA, "Q C-y", "▽test message", "", SKK_INPUT_MODE_HANKAKU_KATAKANA },
+    { SKK_INPUT_MODE_LATIN, "C-y", "", "test message", SKK_INPUT_MODE_LATIN },
+    { SKK_INPUT_MODE_WIDE_LATIN, "C-y", "", "test message", SKK_INPUT_MODE_WIDE_LATIN },
+    { 0, NULL }
+  };
+
+  rule = skk_rule_new ("test-selection", &error);
+  g_assert_no_error(error);
+  skk_context_set_typing_rule(context, rule);
+  g_object_unref(rule);
+  check_transitions (context, transitions);
+  destroy_context(context);
+}
+
 static void
 start_preedit_no_delete (void) {
   SkkContext *context;
@@ -479,6 +586,8 @@ abort_to_latin_commands (void) {
     { SKK_INPUT_MODE_HIRAGANA, "P o p o", "▽ぽぽ", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "P o p o SPC C-l", "▽ぽぽ", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "P o p o C-l", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "k y C-g", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "k y C-l", "", "", SKK_INPUT_MODE_HIRAGANA },
     // abort-to-latin-unhandled: Test cases with no discarded inputs.
     // In these cases, input mode should be changed to latin mode.
     // Note: While these tests cannot represent, the key event will be
@@ -487,13 +596,14 @@ abort_to_latin_commands (void) {
     { SKK_INPUT_MODE_HIRAGANA, "Q", "", "", SKK_INPUT_MODE_LATIN },
     { SKK_INPUT_MODE_HIRAGANA, "a Q", "", "あ", SKK_INPUT_MODE_LATIN },
     // abort-to-latin-unhandled: Test cases with discarded inputs.
-    // These should be exactly the same behaviour as `abort-to-latin`.
+    // These should be exactly the same behaviour as `abort-to-latin` and `abort`.
     { SKK_INPUT_MODE_HIRAGANA, "A Q", "", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "P Q", "", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "P o p Q", "", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "P o p o", "▽ぽぽ", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "P o p o SPC Q", "▽ぽぽ", "", SKK_INPUT_MODE_HIRAGANA },
     { SKK_INPUT_MODE_HIRAGANA, "P o p o Q", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "k y Q", "", "", SKK_INPUT_MODE_HIRAGANA },
     { 0, NULL }
   };
 
@@ -507,6 +617,38 @@ abort_to_latin_commands (void) {
   destroy_context (context);
 }
 
+static void
+commit_unhandled_with_incomplete_kana (void)
+{
+  SkkContext *context;
+  SkkTransition transitions[] = {
+    { SKK_INPUT_MODE_HIRAGANA, "t RET", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { SKK_INPUT_MODE_HIRAGANA, "T RET", "", "", SKK_INPUT_MODE_HIRAGANA },
+    { 0, NULL }
+  };
+  gint retval;
+
+  context = create_context (TRUE, TRUE);
+  check_transitions (context, transitions);
+
+  retval = skk_context_process_key_events (context, "t");
+  g_assert (retval);
+  retval = skk_context_process_key_events (context, "RET");
+  // `commit-unhandled` command here should let RET pass-through
+  // since it does not commit anything.
+  g_assert (!retval);
+
+  skk_context_reset (context);
+  retval = skk_context_process_key_events (context, "T");
+  g_assert (retval);
+  retval = skk_context_process_key_events (context, "RET");
+  // `commit-unhandled` command here should let RET pass-through
+  // since it does not commit anything.
+  g_assert (!retval);
+
+  destroy_context (context);
+}
+
 int
 main (int argc, char **argv) {
   skk_init ();
@@ -550,10 +692,19 @@ main (int argc, char **argv) {
   g_test_add ("/libskk/numeric",
               SkkFixture, numeric_transitions,
               context_setup, test_transitions, context_teardown);
+  g_test_add ("/libskk/no-default-insert-for-mods",
+              SkkFixture, mods_deactivate_default_insert_transitions,
+              context_setup, test_transitions, context_teardown);
+  g_test_add ("/libskk/return-key-preedit",
+              SkkFixture, commit_preedit_transitions,
+              context_setup, test_transitions, context_teardown);
   g_test_add_func ("/libskk/candidate-list", candidate_list);
   g_test_add_func ("/libskk/surrounding", surrounding);
+  g_test_add_func ("/libskk/selection", selection);
   g_test_add_func ("/libskk/start_preedit_no_delete", start_preedit_no_delete);
   g_test_add_func ("/libskk/inherit_typing_rule_for_dict_edit", inherit_typing_rule_for_dict_edit);
   g_test_add_func ("/libskk/abort_to_latin_commands", abort_to_latin_commands);
+  g_test_add_func ("/libskk/commit-unhandled-with-incomplete-kana",
+                   commit_unhandled_with_incomplete_kana);
   return g_test_run ();
 }
